from cmath import isnan
from distutils.log import debug
from enum import Enum
from select import select
import grpc
from json import dumps, loads
import os
from robomotion.utils import File
from robomotion import runner_pb2, runner_pb2_grpc


class AttachConfig:
    def __init__(self, protocol: str, addr: str, pid: int, namespace: str):
        self.protocol = protocol
        self.addr = addr
        self.pid = pid
        self.namespace = namespace


class EProtocol(Enum):
    ProtocolInvalid = ""
    ProtocolNetRPC = "netrpc"
    ProtocolGRPC = "grpc"


class Debug:
    atteched_to: str = ""

    @staticmethod
    def attach(g_addr: str, ns: str):
        cfg = AttachConfig(str(EProtocol.ProtocolGRPC.value), g_addr, os.getpid(), ns)
        cfg_data = dumps(cfg.__dict__)

        Debug.atteched_to = Debug.get_rpc_addr()
        if Debug.atteched_to == "":
            print("empty gRPC address")
            exit(0)

        print("Attached to %s" % Debug.atteched_to)

        channel = grpc.insecure_channel(Debug.atteched_to)
        client = runner_pb2_grpc.DebugStub(channel)

        request = runner_pb2.AttachRequest(config=cfg_data.encode())
        client.Attach(request)
        channel.close()

    @staticmethod
    def detach(ns: str):
        if Debug.atteched_to == "":
            print("empty gRPC address")
            exit(0)

        channel = grpc.insecure_channel(Debug.atteched_to)
        client = runner_pb2_grpc.DebugStub(channel)

        request = runner_pb2.DetachRequest(namespace=ns)
        client.Detach(request)
        channel.close()

    @staticmethod
    def get_rpc_addr() -> str:
        # Import psutil only when needed to avoid macOS permission errors
        from psutil._common import CONN_LISTEN
        
        tabs = Debug.get_netstat_ports(CONN_LISTEN, "robomotion-runner")
        tabs = Debug.filter_tabs(tabs)

        if len(tabs) == 0:
            return ""
        elif len(tabs) == 1:
            return "%s:%d" % (tabs[0]["sconn"].laddr.ip, tabs[0]["sconn"].laddr.port)

        return Debug.select_tab(tabs)

    @staticmethod
    def filter_tabs(tabs: list) -> list:
        filtered = []
        for tab in tabs:
            addr = "%s:%d" % (tab["sconn"].laddr.ip, tab["sconn"].laddr.port)
            try:
                tab["robot_name"] = Debug.get_robot_name(addr)
                filtered.append(tab)
            except Exception as e:
                print(e)

        return filtered

    @staticmethod
    def get_netstat_ports(state: str, process_name: str = ""):
        # Import psutil only when needed to avoid macOS permission errors
        from psutil import net_connections, process_iter
        
        try:
            tabs = net_connections()
        except PermissionError as e:
            print(f"Permission denied accessing network connections. On macOS, you may need to run with elevated privileges or grant permission in System Preferences > Security & Privacy > Privacy > Full Disk Access.")
            print(f"Error: {e}")
            return []

        pids = []
        try:
            for proc in process_iter():
                if process_name.lower() in proc.name().lower():
                    pids.append(proc.pid)
        except Exception as e:
            print(f"Error accessing process information: {e}")
            return []

        filtered = filter(lambda tab: tab.status == state and tab.pid in pids, tabs)
        return list(map(lambda tab: {"sconn": tab}, filtered))

    @staticmethod
    def select_tab(tabs: list) -> str:
        count = len(tabs)

        robots = ""
        for i in range(0, count):
            robots += "%d) %s\n" % (i + 1, tabs[i]["robot_name"])

        selected = 0
        print("Found %d robots running on the machine:" % count)
        print("%s" % robots)
        print("Please select a robot to attach (1-%d):" % count)

        while True:
            line = input()
            if not line.isnumeric():
                continue

            selected = int(line)
            if selected > 0 and selected <= count:
                tab = tabs[selected - 1]
                return "%s:%d" % (tab["sconn"].laddr.ip, tab["sconn"].laddr.port)

    @staticmethod
    def get_robot_name(addr: str) -> str:
        channel = grpc.insecure_channel(addr)
        client = runner_pb2_grpc.RunnerStub(channel)

        request = runner_pb2.Null()
        resp = client.RobotName(request)
        channel.close()
        return resp.robot_name



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from robomotion import runner_pb2 as runner__pb2

GRPC_GENERATED_VERSION = '1.72.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

# try:
#     from grpc._utilities import first_version_is_lower
#     _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
# except ImportError:
#     _version_not_supported = True
# 
# if _version_not_supported:
#     raise RuntimeError(
#         f'The grpc package installed is at version {GRPC_VERSION},'
#         + f' but the generated code in runner_pb2_grpc.py depends on'
#         + f' grpcio>={GRPC_GENERATED_VERSION}.'
#         + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
#         + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
#     )


class RunnerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Init = channel.unary_unary(
                '/proto.Runner/Init',
                request_serializer=runner__pb2.InitRunnerRequest.SerializeToString,
                response_deserializer=runner__pb2.Null.FromString,
                _registered_method=True)
        self.Run = channel.unary_unary(
                '/proto.Runner/Run',
                request_serializer=runner__pb2.RunRequest.SerializeToString,
                response_deserializer=runner__pb2.Null.FromString,
                _registered_method=True)
        self.Clear = channel.unary_unary(
                '/proto.Runner/Clear',
                request_serializer=runner__pb2.Null.SerializeToString,
                response_deserializer=runner__pb2.Null.FromString,
                _registered_method=True)
        self.RobotName = channel.unary_unary(
                '/proto.Runner/RobotName',
                request_serializer=runner__pb2.Null.SerializeToString,
                response_deserializer=runner__pb2.RobotNameResponse.FromString,
                _registered_method=True)


class RunnerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Init(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Run(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Clear(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RobotName(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RunnerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Init': grpc.unary_unary_rpc_method_handler(
                    servicer.Init,
                    request_deserializer=runner__pb2.InitRunnerRequest.FromString,
                    response_serializer=runner__pb2.Null.SerializeToString,
            ),
            'Run': grpc.unary_unary_rpc_method_handler(
                    servicer.Run,
                    request_deserializer=runner__pb2.RunRequest.FromString,
                    response_serializer=runner__pb2.Null.SerializeToString,
            ),
            'Clear': grpc.unary_unary_rpc_method_handler(
                    servicer.Clear,
                    request_deserializer=runner__pb2.Null.FromString,
                    response_serializer=runner__pb2.Null.SerializeToString,
            ),
            'RobotName': grpc.unary_unary_rpc_method_handler(
                    servicer.RobotName,
                    request_deserializer=runner__pb2.Null.FromString,
                    response_serializer=runner__pb2.RobotNameResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'proto.Runner', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('proto.Runner', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Runner(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Init(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Runner/Init',
            runner__pb2.InitRunnerRequest.SerializeToString,
            runner__pb2.Null.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Run(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Runner/Run',
            runner__pb2.RunRequest.SerializeToString,
            runner__pb2.Null.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Clear(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Runner/Clear',
            runner__pb2.Null.SerializeToString,
            runner__pb2.Null.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RobotName(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Runner/RobotName',
            runner__pb2.Null.SerializeToString,
            runner__pb2.RobotNameResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class RobotHelperStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """


class RobotHelperServicer(object):
    """Missing associated documentation comment in .proto file."""


def add_RobotHelperServicer_to_server(servicer, server):
    rpc_method_handlers = {
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'proto.RobotHelper', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('proto.RobotHelper', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class RobotHelper(object):
    """Missing associated documentation comment in .proto file."""


class DebugStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Attach = channel.unary_unary(
                '/proto.Debug/Attach',
                request_serializer=runner__pb2.AttachRequest.SerializeToString,
                response_deserializer=runner__pb2.Nil.FromString,
                _registered_method=True)
        self.Detach = channel.unary_unary(
                '/proto.Debug/Detach',
                request_serializer=runner__pb2.DetachRequest.SerializeToString,
                response_deserializer=runner__pb2.Nil.FromString,
                _registered_method=True)


class DebugServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Attach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Detach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DebugServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Attach': grpc.unary_unary_rpc_method_handler(
                    servicer.Attach,
                    request_deserializer=runner__pb2.AttachRequest.FromString,
                    response_serializer=runner__pb2.Nil.SerializeToString,
            ),
            'Detach': grpc.unary_unary_rpc_method_handler(
                    servicer.Detach,
                    request_deserializer=runner__pb2.DetachRequest.FromString,
                    response_serializer=runner__pb2.Nil.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'proto.Debug', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('proto.Debug', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Debug(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Attach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Debug/Attach',
            runner__pb2.AttachRequest.SerializeToString,
            runner__pb2.Nil.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Detach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Debug/Detach',
            runner__pb2.DetachRequest.SerializeToString,
            runner__pb2.Nil.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from robomotion import health_pb2 as health__pb2

GRPC_GENERATED_VERSION = '1.72.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

# try:
#     from grpc._utilities import first_version_is_lower
#     _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
# except ImportError:
#     _version_not_supported = True
# 
# if _version_not_supported:
#     raise RuntimeError(
#         f'The grpc package installed is at version {GRPC_VERSION},'
#         + f' but the generated code in health_pb2_grpc.py depends on'
#         + f' grpcio>={GRPC_GENERATED_VERSION}.'
#         + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
#         + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
#     )


class HealthStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Check = channel.unary_unary(
                '/grpc.health.v1.Health/Check',
                request_serializer=health__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=health__pb2.HealthCheckResponse.FromString,
                _registered_method=True)


class HealthServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Check(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HealthServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Check': grpc.unary_unary_rpc_method_handler(
                    servicer.Check,
                    request_deserializer=health__pb2.HealthCheckRequest.FromString,
                    response_serializer=health__pb2.HealthCheckResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'grpc.health.v1.Health', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('grpc.health.v1.Health', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Health(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Check(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/grpc.health.v1.Health/Check',
            health__pb2.HealthCheckRequest.SerializeToString,
            health__pb2.HealthCheckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: health.proto
# Protobuf Python Version: 6.30.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# Comment out version validation to avoid version mismatch error
# _runtime_version.ValidateProtobufRuntimeVersion(
#     _runtime_version.Domain.PUBLIC,
#     6,
#     30,
#     0,
#     '',
#     'health.proto'
# )
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0chealth.proto\x12\x0egrpc.health.v1\"%\n\x12HealthCheckRequest\x12\x0f\n\x07service\x18\x01 \x01(\t\"\x94\x01\n\x13HealthCheckResponse\x12\x41\n\x06status\x18\x01 \x01(\x0e\x32\x31.grpc.health.v1.HealthCheckResponse.ServingStatus\":\n\rServingStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SERVING\x10\x01\x12\x0f\n\x0bNOT_SERVING\x10\x02\x32Z\n\x06Health\x12P\n\x05\x43heck\x12\".grpc.health.v1.HealthCheckRequest\x1a#.grpc.health.v1.HealthCheckResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'health_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_HEALTHCHECKREQUEST']._serialized_start=32
  _globals['_HEALTHCHECKREQUEST']._serialized_end=69
  _globals['_HEALTHCHECKRESPONSE']._serialized_start=72
  _globals['_HEALTHCHECKRESPONSE']._serialized_end=220
  _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_start=162
  _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_end=220
  _globals['_HEALTH']._serialized_start=222
  _globals['_HEALTH']._serialized_end=312
# @@protoc_insertion_point(module_scope)



import grpc
import asyncio
import time

from robomotion import plugin_pb2
from robomotion import plugin_pb2_grpc
from robomotion.error import Error
from robomotion.message import Context, Message
from robomotion.runtime import Runtime
from robomotion.capabilities import get_capabilities

class Node:
    def __init__(self):
        self.guid = ''
        self.name = ''
        self.delayBefore = 0.0
        self.delayAfter = 0.0
        self.continueOnError = False
        self.scope = ''

    def get_var_name(self, var):
        d = {v: k for k, v in globals().items()}
        return d[var]

    async def on_create(self):
        pass

    async def on_message(self, ctx: Context):
        pass

    async def on_close(self):
        pass
    def get_capabilities(self):
        pass


class NodeServicer(plugin_pb2_grpc.NodeServicer):
    def Init(self, request: plugin_pb2.InitRequest, context):
        try:
            channel = grpc.insecure_channel('127.0.0.1:%d' % request.port)
            client = plugin_pb2_grpc.RuntimeHelperStub(channel)
            Runtime.set_client(client=client)
            Runtime.check_runner_conn(channel)

            return plugin_pb2.Empty()

        except Exception as e:
            context.set_code(grpc.StatusCode.UNKNOWN)
            if isinstance(e, Error):
                context.set_details(e.serialize())
            else:
                context.set_details(Error('Err.Unknown', str(e)).serialize())

    def OnCreate(self, request: plugin_pb2.OnCreateRequest, context):
        try:
            Runtime.active_nodes += 1

            config = request.config
            factory = Runtime.factories[request.name]
            if factory:
                asyncio.run(factory.on_create(config))

            return plugin_pb2.OnCreateResponse()

        except Exception as e:
            context.set_code(grpc.StatusCode.UNKNOWN)
            if isinstance(e, Error):
                context.set_details(e.serialize())
            else:
                context.set_details(Error('Err.Unknown', str(e)).serialize())

    def OnMessage(self, request: plugin_pb2.OnMessageRequest, context):
        node = Runtime.nodes[request.guid]
        try:
            data = Runtime.decompress(request.inMessage)

            if request.guid in Runtime.nodes:
                # Sleep with delayBefore on enter
                if node.delayBefore > 0:
                    time.sleep(node.delayBefore)

                ctx = Message(data)
                asyncio.run(node.on_message(ctx))
                out_message = ctx.get_raw()

                # Sleep with delayAfter on exit
                if node.delayAfter > 0:
                    time.sleep(node.delayAfter)

                return plugin_pb2.OnMessageResponse(outMessage=out_message)

            return plugin_pb2.OnMessageResponse(outMessage=data)

        except Exception as e:
            # Print the full exception with traceback for debugging
            import traceback
            print(f"Exception in OnMessage for node {request.guid}: {e}")
            traceback.print_exc()

            if node.continueOnError:
                return plugin_pb2.OnMessageResponse(outMessage=data)

            context.set_code(grpc.StatusCode.UNKNOWN)
            if isinstance(e, Error):
                context.set_details(e.serialize())
            else:
                context.set_details(Error('Err.Unknown', str(e)).serialize())

            # Always return a valid response to avoid grpc serialization errors
            return plugin_pb2.OnMessageResponse(outMessage=data)

    def OnClose(self, request: plugin_pb2.OnCloseRequest, context):
        try:
            if request.guid in Runtime.nodes:
                node = Runtime.nodes[request.guid]
                if node:
                    asyncio.run(node.on_close())

            Runtime.active_nodes -= 1
            if Runtime.active_nodes <= 0:
                Runtime.event.set()

            return plugin_pb2.OnCloseResponse()

        except Exception as e:
            context.set_code(grpc.StatusCode.UNKNOWN)
            if isinstance(e, Error):
                context.set_details(e.serialize())
            else:
                context.set_details(Error('Err.Unknown', str(e)).serialize())
    def GetCapabilities(self, request: plugin_pb2.Empty, context):        
        return plugin_pb2.PGetCapabilitiesResponse(capabilities=get_capabilities())
          
